// Mockingboard emulation module
//
// Tom Charlesworth
//
// History:
//
// v1.12.03 (11 Jan 2004)
// - For free-running 6522 timer1 IRQ, reload with current ACCESS_TIMER1 value.
//   (Fixes Ultima 4/5 playback speed problem.)
//
// v1.12.01 (24 Nov 2002)
// - Shaped the tone waveform more logarithmicly
// - Added support for MB ena/dis switch on Config dialog
// - Added log file support
//
// v1.12.00 (17 Nov 2002)
// - Initial version (no AY8910 envelope support)
//

#include <dsound.h>
#include <crtdbg.h>
#include "stdhdr.h"
#pragma  hdrstop
#include "applewin.h"

#define SY6522_DEVICE_A 0
#define SY6522_DEVICE_B 1

#define SLOT4 4
#define SLOT5 5

#define NUM_MB 2
#define NUM_DEVS_PER_MB 2
#define NUM_VOICES_PER_AY8910 3
#define NUM_VOICES (NUM_MB*NUM_DEVS_PER_MB*NUM_VOICES_PER_AY8910)

#define NUM_NOISES (NUM_MB*NUM_DEVS_PER_MB)
#define NOISE_BASE_VOICE_NUMBER NUM_VOICES

#define NUM_TOTAL_VOICES (NUM_VOICES+NUM_NOISES)

// Chip offsets from card base.
#define SY6522A_Offset	0x00
#define SY6522B_Offset	0x80
#define Votrax_Offset	0x40


// 6522 registers:
typedef struct
{
	BYTE ORB;				// $00 - Port B
	BYTE ORA;				// $01 - Port A (with handshaking)
	BYTE DDRB;				// $02 - Data Direction Register B
	BYTE DDRA;				// $03 - Data Direction Register A
	BYTE TIMER1L;			// $04 - Read / Write & latch.
	BYTE TIMER1H;			// $05 - Read / Write & initiate count.
	BYTE ACCESS_TIMER1L;	// $06
	BYTE ACCESS_TIMER1H;	// $07
	BYTE TIMER2L;			// $08
	BYTE ACCESS_TIMER2H;	// $09
	BYTE SERIAL_SHIFT;		// $0A
	BYTE ACR;				// $0B - Auxiliary Control Register
	BYTE PCR;				// $0C - Peripheral Control Register
	BYTE IFR;				// $0D - Interrupt Flag Register
	BYTE IER;				// $0E - Interrupt Enable Register
	BYTE ORA_NO_HS;			// $0F - Port A (without handshaking)
} SY6522;

// AY-3-8910 registers:
typedef struct
{
	BYTE A_Tone_Fine;
	BYTE A_Tone_Coarse;
	BYTE B_Tone_Fine;
	BYTE B_Tone_Coarse;
	BYTE C_Tone_Fine;
	BYTE C_Tone_Coarse;
	BYTE Noise;
	BYTE Enable;
	BYTE A_Amplitude;
	BYTE B_Amplitude;
	BYTE C_Amplitude;
	BYTE Envelope_Fine;
	BYTE Envelope_Coarse;
	BYTE Envelope_Shape;
	BYTE Undefined1;
	BYTE Undefined2;
} AY8910;

typedef struct
{
	SY6522 sy6522;
	AY8910 ay8910;
	BYTE nAY8910Number;
	BYTE nAYCurrentRegister;
	BYTE nTimerStatus;
	BYTE nAYCurrentTIMER1L;	// For Skyfox MB detection
	signed int nNoisePeriodRemaining;
	SHORT nNoiseCurrData;
	USHORT nNoiseLastBit;	// Last output bit from LFSR
} SY6522_AY8910;


// IFR & IER:
#define PERIPHERAL	(1<<1)
#define TIMER2		(1<<5)
#define TIMER1		(1<<6)

// ACR:
#define RUNMODE		(1<<6)	// 0 = 1-Shot Mode, 1 = Free Running Mode
#define RM_ONESHOT		(0<<6)
#define RM_FREERUNNING	(1<<6)


// Votrax registers:
#define VO_DURPHON	0x00
#define VO_INFLECT	0x01
#define VO_RATEINF	0x02
#define VO_CTTRAMP	0x03
#define VO_FILFREQ	0x04


// Support 2 MB's, each with 2x SY6522/AY8910 pairs.
static SY6522_AY8910 g_MB[NUM_MB * NUM_DEVS_PER_MB];

// Global dialog var:
bool g_bMBEnable = true;		// Mockingboard enable flag

// Global timer vars
static ULONG g_n6522TimerPeriod = 0;
static USHORT g_nMBTimerDevice = 0;	// SY6522 device# which is generating timer IRQ


// External global timer vars
bool g_bMBTimerIrqActive = false;

// Prototypes
static bool PlayVoice(int n, DWORD nFreq, LONG nVolume);
static bool StopVoice(int n);


// Debug:

#define DBG_MB_STATS 0

typedef struct
{
	USHORT nMinTonePer;
	USHORT nMaxTonePer;
	USHORT nMinNoisePer;
	USHORT nMaxNoisePer;
	USHORT nMinEnvPer;
	USHORT nMaxEnvPer;
	DWORD nMinFreq;
	DWORD nMaxFreq;
	DWORD nCurrActiveVoices;
	DWORD nMaxActiveVoices;
} MB_STATS;

static MB_STATS MbStats = {0};

//---------------------------------------------------------------------------

#define AY_VOLUME_MAX 0x0F
#define AY_ENVELOPE_MASK 0x10

#define ALL_TONE_MASK 0x07
#define ALL_NOISE_MASK 0x38

#define AY_TONE_PERIOD_MASK 0x0FFF
#define AY_NOISE_PERIOD_MASK 0x1F

#define MIN_TONE_PERIOD 0x10 	// Arbitrary minimum tone period

static LONG aVolTbl[16] = {0};

const double _M14 = 14.31818e6;
const double CLK_6502 = (_M14 / 14.0);

const double MB_CLK_FREQ = (CLK_6502 / 16.0);	// AY8910 Tone & Noise CLK = 6502 CLK / 16

//---------------------------------------------------------------------------

const DWORD nBit0 = 1<<0;
const DWORD nBit2 = 1<<2;

// The LFSR is clocked at a frequency of MB_CLK_FREQ
static USHORT AYNoise()
{
	static DWORD nLFSR = 0x00001;	// 17-bit LFSR
	USHORT nNoise = (USHORT) (nLFSR & 1);
	DWORD n16=0;

	if(nLFSR & nBit2) n16 = 0x10000;
	if(nLFSR & nBit0) n16 ^= 0x10000;
	nLFSR >>= 1;
	nLFSR ^= n16;

	return nNoise;
}

static void DoAY8910(SY6522_AY8910* pMB)
{
	DWORD nAYPeriod,nFreq;
	LONG nVolume;

	for(int nVoice=0,nMask=0x01; nVoice<NUM_VOICES_PER_AY8910; nVoice++,nMask<<=1)
	{
		BYTE nAYVolume = *((&pMB->ay8910.A_Amplitude)+nVoice);	// For tone & noise
		if(nAYVolume & AY_ENVELOPE_MASK)
		{
			nAYVolume = AY_VOLUME_MAX;	// Set to max volume as envelope isn't supported
#if DBG_MB_STATS
			USHORT nAYEnvPeriod = *(USHORT*) (&pMB->ay8910.Envelope_Fine);
			if((nAYEnvPeriod < MbStats.nMinEnvPer) || (MbStats.nMinEnvPer == 0))
				MbStats.nMinEnvPer = nAYEnvPeriod;
			if (nAYEnvPeriod > MbStats.nMaxEnvPer)
				MbStats.nMaxEnvPer = nAYEnvPeriod;
#endif
		}
		nAYVolume &= AY_VOLUME_MAX;
		nVolume = aVolTbl[nAYVolume];

		nAYPeriod = (DWORD) *(USHORT*) ((&pMB->ay8910.A_Tone_Fine)+nVoice*2);
		nAYPeriod &= AY_TONE_PERIOD_MASK;
		if(nAYPeriod == 1) nAYPeriod = 0;
		if((nAYPeriod > 1) && (nAYPeriod <= MIN_TONE_PERIOD))
			nAYPeriod = 0;

		if(!(pMB->ay8910.Enable & nMask))
		{
			if(nAYVolume && nAYPeriod)
			{
				nFreq = (DWORD) (MB_CLK_FREQ / ((double)nAYPeriod / 2.0));
				if(nFreq < DSBFREQUENCY_MIN)
					nFreq = DSBFREQUENCY_MIN;
				if(nFreq > DSBFREQUENCY_MAX)
					nFreq = DSBFREQUENCY_MAX;

#if DBG_MB_STATS
				if((nAYPeriod < MbStats.nMinTonePer) || (MbStats.nMinTonePer == 0))
					MbStats.nMinTonePer = (USHORT) nAYPeriod;
				if (nAYPeriod > MbStats.nMaxTonePer)
					MbStats.nMaxTonePer = (USHORT) nAYPeriod;
#endif
				PlayVoice(pMB->nAY8910Number*NUM_VOICES_PER_AY8910+nVoice, nFreq, nVolume);
			}
			else
			{
				PlayVoice(pMB->nAY8910Number*NUM_VOICES_PER_AY8910+nVoice, DSBFREQUENCY_MIN, DSBVOLUME_MIN);
			}
		}
		else
		{
			StopVoice(pMB->nAY8910Number*NUM_VOICES_PER_AY8910+nVoice);
		}
	}

	//

	int nNoiseVoice = NOISE_BASE_VOICE_NUMBER + pMB->nAY8910Number;
	nAYPeriod = pMB->ay8910.Noise & AY_NOISE_PERIOD_MASK;

	if(((pMB->ay8910.Enable & ALL_NOISE_MASK) != ALL_NOISE_MASK) && nAYPeriod)
	{
		PlayVoice(nNoiseVoice, 0, 0);
	}
	else
	{
		StopVoice(nNoiseVoice);
	}
}


static void StartTimer(SY6522_AY8910* pMB)
{
	if((pMB->nAY8910Number & 1) != SY6522_DEVICE_A)
		return;

	if((pMB->sy6522.IER & TIMER1) == 0x00)
		return;

	USHORT nPeriod = *(USHORT*)&pMB->sy6522.TIMER1L;

	if(nPeriod < 0x4000)	// Bit of IRQ protection (probably not needed)
	{
		if(nPeriod <= 0xff)	// Timer1L value has been written (but not Timer1H yet)
			return;

		_ASSERT(0);
		return;
	}

	pMB->nTimerStatus = 1;

	// 6522 CLK runs at same speed as 6502 CLK
	g_n6522TimerPeriod = nPeriod;

	g_bMBTimerIrqActive = true;
	g_nMBTimerDevice = pMB->nAY8910Number;
}

static void StopTimer(SY6522_AY8910* pMB)
{
	pMB->nTimerStatus = 0;
	g_n6522TimerPeriod = 0;
	g_bMBTimerIrqActive = false;
	g_nMBTimerDevice = 0;
}


static void ResetSY6522(SY6522_AY8910* pMB)
{
	memset(&pMB->sy6522,0,sizeof(SY6522));

	if(pMB->nTimerStatus)
		StopTimer(pMB);

	pMB->nAYCurrentRegister = 0;
}

static void ResetAY8910(SY6522_AY8910* pMB)
{
	memset(&pMB->ay8910,0,sizeof(AY8910));
	pMB->ay8910.Enable = 0x3F;	// Disable all noise & voices

	DoAY8910(pMB);	// Stop any active voices
}


static void SY6522_Write(BYTE nDevice, BYTE nReg, BYTE nValue)
{
	SY6522_AY8910* pMB = &g_MB[nDevice];
	BYTE nAYReg;

	switch (nReg)
	{
		case 0x00:	// ORB
			nValue &= pMB->sy6522.DDRB;

			// Ultima pads values written to 6522 with 1's, so mask these
			// out to get true value.
			nValue &= 0x07;		// Only lower 3 data lines used.

			if(nValue == 0x04)
			{
				nValue = pMB->sy6522.ORB;	// Get last value written
				switch (nValue)
				{
					case 7:
						// LATCH: Select value in Port A as current AY-3-8910 register
						pMB->nAYCurrentRegister = pMB->sy6522.ORA;
						break;
					case 6:
						// WRITE: Write data in Port A to current AY-3-8910 register
						nAYReg = pMB->nAYCurrentRegister;
						((BYTE*)&pMB->ay8910)[nAYReg] = pMB->sy6522.ORA;
						DoAY8910(pMB);
						pMB->sy6522.ORB = 0x04;	// STORE
						break;
//					case 5:	// Maybe used to read AY regs? Will need to set DDRA to input
//						pMB->sy6522.ORB = 0x04;	// STORE
//						break;
					case 0:
						// RESET: Reset AY8910 only
						ResetAY8910(pMB);
						// And drop into default...
					default:
						// Ultima writes (after masking with %00000111):
						// 04,00,04 to ORB: to RESET.
						// 04,07,04 to ORB: to LATCH an AY8910 register.
						// 06,04,04 to ORB: to WRITE to AY8910 register.
						pMB->sy6522.ORB = 0x04;	// STORE
						break;
				}
			}
			else
			{
				pMB->sy6522.ORB = nValue;	// STORE
			}
			break;
		case 0x01:	// ORA
			pMB->sy6522.ORA = nValue & pMB->sy6522.DDRA;
			break;
		case 0x02:	// DDRB
			pMB->sy6522.DDRB = nValue;
			break;
		case 0x03:	// DDRA
			pMB->sy6522.DDRA = nValue;
			break;
		case 0x04:	// TIMER1L
		case 0x06:	// ACCESS_TIMER1L
			pMB->sy6522.ACCESS_TIMER1L = nValue;
			break;
		case 0x05:	// TIMER1H
			/* Initiates timer & Clears time-out of timer1 */

			// Clear Timer Interrupt Flag.
			pMB->sy6522.IFR &= ~TIMER1;

			pMB->sy6522.ACCESS_TIMER1H = nValue;
			pMB->sy6522.TIMER1H = nValue;
			pMB->sy6522.TIMER1L = pMB->sy6522.ACCESS_TIMER1L;

			StartTimer(pMB);
			break;
		case 0x07:	// ACCESS_TIMER1H
			// Clear Timer Interrupt Flag.
			pMB->sy6522.ACCESS_TIMER1H = nValue;
			pMB->sy6522.IFR &= ~TIMER1;
			break;
		case 0x08:	// TIMER2L
			break;
		case 0x09:	// ACCESS_TIMER2H
			break;
		case 0x0a:	// SERIAL_SHIFT
			break;
		case 0x0b:	// ACR
			pMB->sy6522.ACR = nValue;
			break;
		case 0x0c:	// PCR -  Used for Speech chip only
			pMB->sy6522.PCR = nValue;
			break;
		case 0x0d:	// IFR
			// - Clear those bits which are set in the lower 7 bits.
			// - Can't clear bit 7.
			nValue |= 0x80;	// Set high bit
			nValue ^= 0x7F;	// Make mask
			pMB->sy6522.IFR &= nValue;
			break;
		case 0x0e:	// IER
			if(!(nValue & 0x80))
			{
				// Clear those bits which are set in the lower 7 bits.
				nValue ^= 0x7F;
				pMB->sy6522.IER &= nValue;
	
				// Check if timer has been disabled.
				if(pMB->sy6522.IER & TIMER1)
					break;

				if(pMB->nTimerStatus == 0)
					break;
				
				pMB->nTimerStatus = 0;
				
				// Stop timer
				StopTimer(pMB);
			}
			else
			{
				// Set those bits which are set in the lower 7 bits.
				nValue &= 0x7F;
				pMB->sy6522.IER |= nValue;
				StartTimer(pMB);
			}
			break;
		case 0x0f:	// ORA_NO_HS
			break;
	}
}

static BYTE SY6522_Read(BYTE nDevice, BYTE nReg)
{
	SY6522_AY8910* pMB = &g_MB[nDevice];
	BYTE nValue = 0x00;

	switch (nReg)
	{
		case 0x00:	// ORB
			nValue = pMB->sy6522.ORB;
			break;
		case 0x01:	// ORA
			nValue = pMB->sy6522.ORA;
			break;
		case 0x02:	// DDRB
			nValue = pMB->sy6522.DDRB;
			break;
		case 0x03:	// DDRA
			nValue = pMB->sy6522.DDRA;
			break;
		case 0x04:	// TIMER1L
			nValue = pMB->nAYCurrentTIMER1L;
			pMB->nAYCurrentTIMER1L -= 8;	// For Skyfox MB detection routine
			pMB->sy6522.IFR &= ~TIMER1;		// Also clears Timer Interrupt Flag
			break;
		case 0x05:	// TIMER1H
			break;
		case 0x06:	// ACCESS_TIMER1L
			break;
		case 0x07:	// ACCESS_TIMER1H
			break;
		case 0x08:	// TIMER2L
			break;
		case 0x09:	// ACCESS_TIMER2H
			break;
		case 0x0a:	// SERIAL_SHIFT
			break;
		case 0x0b:	// ACR
			nValue = pMB->sy6522.ACR;
			break;
		case 0x0c:	// PCR
			nValue = pMB->sy6522.PCR;
			break;
		case 0x0d:	// IFR
			nValue = pMB->sy6522.IFR;
			break;
		case 0x0e:	// IER
			nValue = pMB->sy6522.IER;
			break;
		case 0x0f:	// ORA_NO_HS
			break;
	}

	return nValue;
}

//===========================================================================

#define NEW_TONE_WAVEFORM 1

#if NEW_TONE_WAVEFORM
const int nHalfPeriodSize = 4;
//const int nHalfPeriodSize = 2;
#endif


#define SAFE_RELEASE(p)      { if(p) { (p)->Release(); (p)=NULL; } }

#define MAX_SOUND_DEVICES 10

static char *sound_devices[MAX_SOUND_DEVICES];
static GUID sound_device_guid[MAX_SOUND_DEVICES];
static int num_sound_devices = 0;
static const unsigned short g_nNumChannels = 2;

static LPDIRECTSOUND g_lpDS = NULL;
static bool g_bDSAvailable = false;

static const DWORD g_dwDSBufferSize = 8192;

static const SHORT nWaveData0 = (SHORT)0x8000;
static const SHORT nWaveData1 = (SHORT)0x7FFF;

static const int g_nNumEvents = NUM_NOISES*2+1;

static HANDLE g_hEvents[g_nNumEvents] = {0};
static HANDLE g_hThread = NULL;

static const int MB_CLK_DIV = 2;

// External globals:
ULONGLONG g_nPerfFreq = 0;


typedef struct
{
	LPDIRECTSOUNDBUFFER lpDSBvoice;
	LPDIRECTSOUNDNOTIFY lpDSNotify;
	bool bActive;
	DWORD nFreq;
	LONG nVolume;
} MB_VOICE;

static MB_VOICE aVoice[NUM_TOTAL_VOICES] = {0};


static BOOL CALLBACK DSEnumProc(LPGUID lpGUID, LPCTSTR lpszDesc, LPCTSTR lpszDrvName,  LPVOID lpContext)
{
	int i = num_sound_devices;
	if(i == MAX_SOUND_DEVICES)
		return TRUE;
	if(lpGUID != NULL)
		memcpy(&sound_device_guid[i], lpGUID, sizeof (GUID));
	sound_devices[i] = strdup(lpszDesc);

	if(g_fh) fprintf(g_fh, "%d: %s - %s\n",i,lpszDesc,lpszDrvName);

	num_sound_devices++;
	return TRUE;
}


#ifdef _DEBUG
static char *DirectSound_ErrorText (HRESULT error)
{
    switch( error )
    {
    case DSERR_ALLOCATED:
        return "Allocated";
    case DSERR_CONTROLUNAVAIL:
        return "Control Unavailable";
    case DSERR_INVALIDPARAM:
        return "Invalid Parameter";
    case DSERR_INVALIDCALL:
        return "Invalid Call";
    case DSERR_GENERIC:
        return "Generic";
    case DSERR_PRIOLEVELNEEDED:
        return "Priority Level Needed";
    case DSERR_OUTOFMEMORY:
        return "Out of Memory";
    case DSERR_BADFORMAT:
        return "Bad Format";
    case DSERR_UNSUPPORTED:
        return "Unsupported";
    case DSERR_NODRIVER:
        return "No Driver";
    case DSERR_ALREADYINITIALIZED:
        return "Already Initialized";
    case DSERR_NOAGGREGATION:
        return "No Aggregation";
    case DSERR_BUFFERLOST:
        return "Buffer Lost";
    case DSERR_OTHERAPPHASPRIO:
        return "Other Application Has Priority";
    case DSERR_UNINITIALIZED:
        return "Uninitialized";
    case DSERR_NOINTERFACE:
        return "No Interface";
    default:
        return "Unknown";
    }
}
#endif


static bool DSGetLock(LPDIRECTSOUNDBUFFER pVoice, DWORD dwOffset, DWORD dwBytes,
					  SHORT** ppDSLockedBuffer0, DWORD* pdwDSLockedBufferSize0,
					  SHORT** ppDSLockedBuffer1, DWORD* pdwDSLockedBufferSize1)
{
	DWORD nStatus;
	HRESULT hr = pVoice->GetStatus(&nStatus);
	if(hr != DS_OK)
		return false;

	if(nStatus & DSBSTATUS_BUFFERLOST)
	{
		do
		{
			hr = pVoice->Restore();
			if(hr == DSERR_BUFFERLOST)
				Sleep(10);
		}
		while(hr != DS_OK);
	}

	// Get write only pointer(s) to sound buffer
	if(dwBytes == 0)
	{
		if(FAILED(hr = pVoice->Lock(0, 0,
								(void**)ppDSLockedBuffer0, pdwDSLockedBufferSize0,
								(void**)ppDSLockedBuffer1, pdwDSLockedBufferSize1,
								DSBLOCK_ENTIREBUFFER)))
			return false;
	}
	else
	{
		if(FAILED(hr = pVoice->Lock(dwOffset, dwBytes,
								(void**)ppDSLockedBuffer0, pdwDSLockedBufferSize0,
								(void**)ppDSLockedBuffer1, pdwDSLockedBufferSize1,
								0)))
			return false;
	}

	return true;
}


#if NEW_TONE_WAVEFORM
SHORT ToneFunc(int n)
{
	// y(x) = -0.0145x4 + 2.4764x3 - 166.2x2 + 5255x (x <- [0..45])
	//
	// x                     y(n)
	// ----------------------------------
	// 0                     y(0)
	// 1                     y(1)
	// 2                     y(2)
	// ...
	// nHalfPeriodSize-1     y(nHalfPeriodSize-1)
	// nHalfPeriodSize       y(nHalfPeriodSize)
	// nHalfPeriodSize+1    -y(1)
	// nHalfPeriodSize+2    -y(2)
	// ...
	// nHalfPeriodSize*2-1  -y(nHalfPeriodSize-1)
	//
	// y(0)=0=0x8000, y(nHalfPeriodSize)=65535=0xFFFF
	// Adjust and return: y(0)=-32768=0x8000, y(nHalfPeriodSize)=32767=0x7FFF
	//
	// # samples = (1+nHalfWaveSize) + (nHalfWaveSize-1) = 2 x nHalfWaveSize

	n %= nHalfPeriodSize*2;				// Offset in complete wave

	int n1 = n % (nHalfPeriodSize+1);	// Offset in 1st half of wave
	if(n >= (nHalfPeriodSize+1)) n1++;	// Skip repeated value at apex of curve

	// Return clamped boundary values
	if(n==0) return nWaveData0;
	if(n==nHalfPeriodSize) return nWaveData1;

	const double scale = 45.0 / (double)nHalfPeriodSize;
	double x = (double)n1 * scale;

	const double c4 = -0.0145;
	const double c3 = 2.4764;
	const double c2 = -166.2;
	const double c1 = 5255.0;

	double y = c4*pow(x,4) + c3*pow(x,3) + c2*pow(x,2) + c1*x;

	// Clamp value to unsigned range [0x0000,..,0xFFFF]
	if(y < 0.0) y = 0.0;
	if(y > 65535.0) y = 65535.0;

	// Shift data into signed range [0x8000,..,0x0000,..,0x7FFF]
	if(n < (nHalfPeriodSize+1))
		y = y - (double)(USHORT)nWaveData0;
	else
		y = -(y - (double)(USHORT)nWaveData0);

	// Clamp value to signed range [nWaveData0..nWaveData1]
	if(y < (double)nWaveData0) y = (double)nWaveData0;
	if(y > (double)nWaveData1) y = (double)nWaveData1;

	return (SHORT) y;
}
#else
SHORT ToneFunc(int n)
{
	return (n&1) ? nWaveData1 : nWaveData0;
}
#endif


void DSInitToneBuffer(SHORT* pDSLockedBuffer, int i)
{
	if((i<=2) || ((i>=6) && (i<=8)))
	{
		// Left Speaker:  Voices: 0,1,2; 6,7,8
		for(DWORD j=0; j<g_dwDSBufferSize/4; j++)
		{
			*(pDSLockedBuffer+j*2+0) = ToneFunc(j);
			*(pDSLockedBuffer+j*2+1) = 0;
		}
	}
	else
	{
		// Right Speaker: Voices: 3,4,5; 9,A,B
		for(DWORD j=0; j<g_dwDSBufferSize/4; j++)
		{
			*(pDSLockedBuffer+j*2+0) = 0;
			*(pDSLockedBuffer+j*2+1) = ToneFunc(j);
		}
	}
}


void DSInitNoiseBuffer(SHORT* pDSLockedBuffer, DWORD dwDSLockedBufferSize, int nAyNum)
{
	signed int nPeriod = g_MB[nAyNum].nNoisePeriodRemaining;
	SHORT n = g_MB[nAyNum].nNoiseCurrData;
	USHORT nNoise = g_MB[nAyNum].nNoiseLastBit;

	// Determine which enabled noise channel is the loudest
	UCHAR nAYEnable = g_MB[nAyNum].ay8910.Enable;
	signed char nAYVolume = -1;
	int nVoiceNum = 0;

	if((nAYEnable & ALL_NOISE_MASK) != ALL_NOISE_MASK)
	{
		for(int i=0; i<NUM_VOICES_PER_AY8910; i++)
		{
			signed char nVol = (signed char) (*(&(g_MB[nAyNum].ay8910.A_Amplitude)+i) & 0x1f);
			if(((nAYEnable & (1<<(3+i))) == 0) && (nVol > nAYVolume))
			{
				nVoiceNum = i;
				nAYVolume = nVol;
			}
		}
	}
	else
		nAYVolume = 0;	// No noise enabled

	UCHAR nNoisePeriod = g_MB[nAyNum].ay8910.Noise & AY_NOISE_PERIOD_MASK;

	if(nAYVolume & AY_ENVELOPE_MASK)
		nAYVolume = AY_VOLUME_MAX;
	else
		nAYVolume &= AY_VOLUME_MAX;

	SHORT nVolume = (1<<nAYVolume) - 1;	// [0000..7FFF]

	// Output same noise to both speakers
	while(dwDSLockedBufferSize)
	{
		if(nPeriod <= 0)
		{
			nPeriod += nNoisePeriod;

			if(nNoise != AYNoise())
			{
				if(n & 0x8000)
				{
					n = nVolume;
				}
				else
				{
					if(nVolume==0)
						n = 0;
					else
						n = -nVolume-1;
				}
			}
		}

		*pDSLockedBuffer++ = n;
		*pDSLockedBuffer++ = n;

		nPeriod -= MB_CLK_DIV;
		dwDSLockedBufferSize -= 4;
	}

	g_MB[nAyNum].nNoisePeriodRemaining = nPeriod;
	g_MB[nAyNum].nNoiseCurrData = n;
	g_MB[nAyNum].nNoiseLastBit = nNoise;
}


static VOID ThreadFunction(LPVOID lpParam) 
{
	DWORD dwDSLockedBufferSize0, dwDSLockedBufferSize1;
	SHORT *pDSLockedBuffer0, *pDSLockedBuffer1;
	static DWORD dwDbgLastEvent = -1;

	while(1)
	{
		DWORD dwWaitResult = WaitForMultipleObjects( 
								g_nNumEvents,		// number of handles in array
								g_hEvents,			// array of event handles
								FALSE,				// wait until any one is signaled
								INFINITE);			// wait indefinitely

		if((dwWaitResult < WAIT_OBJECT_0) || (dwWaitResult > WAIT_OBJECT_0+g_nNumEvents-1))
			continue;

		dwWaitResult -= WAIT_OBJECT_0;	// Determine event # that signaled

		if(dwWaitResult == (g_nNumEvents-1))	// Termination event
			break;

		DWORD dwOffset = dwWaitResult & 1 ? g_dwDSBufferSize/2 : 0;
		int nVoice = NUM_VOICES+(dwWaitResult>>1);	// 2 events per voice
		int nNoiseNum = dwWaitResult>>1;

		if(!DSGetLock(aVoice[nVoice].lpDSBvoice,
						dwOffset, g_dwDSBufferSize/2,
						&pDSLockedBuffer0, &dwDSLockedBufferSize0,
						&pDSLockedBuffer1, &dwDSLockedBufferSize1))
			continue;

		// Construct new wave
		_ASSERT(pDSLockedBuffer1 == NULL);
		_ASSERT(dwDSLockedBufferSize0 == (g_dwDSBufferSize/2));

		DSInitNoiseBuffer(pDSLockedBuffer0, dwDSLockedBufferSize0, nNoiseNum);

		// Commit sound buffer
		HRESULT hr = aVoice[nVoice].lpDSBvoice->Unlock((void*)pDSLockedBuffer0, dwDSLockedBufferSize0,
											(void*)pDSLockedBuffer1, dwDSLockedBufferSize1);
	}

	ExitThread(0);
}


static HRESULT DSGetSoundBuffer(MB_VOICE* pVoice, DWORD dwFlags)
{
	WAVEFORMATEX wavfmt;
	DSBUFFERDESC dsbdesc;

	wavfmt.wFormatTag = WAVE_FORMAT_PCM;
	wavfmt.nChannels = g_nNumChannels;
	wavfmt.nSamplesPerSec = (DWORD)(MB_CLK_FREQ/MB_CLK_DIV);	// 31969Hz (for noise only)
	wavfmt.wBitsPerSample = 16;
	wavfmt.nBlockAlign = g_nNumChannels==1 ? 2 : 4;
	wavfmt.nAvgBytesPerSec = wavfmt.nBlockAlign * wavfmt.nSamplesPerSec;

	memset (&dsbdesc, 0, sizeof (dsbdesc));
	dsbdesc.dwSize = sizeof (dsbdesc);
	dsbdesc.dwBufferBytes = g_dwDSBufferSize;
	dsbdesc.lpwfxFormat = &wavfmt;
	dsbdesc.dwFlags = dwFlags | DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_STICKYFOCUS;

	return g_lpDS->CreateSoundBuffer(&dsbdesc, &pVoice->lpDSBvoice, NULL);
}


static bool DSInit()
{
	DWORD dwDSLockedBufferSize = 0;    // Size of the locked DirectSound buffer
	SHORT* pDSLockedBuffer;

	if(g_bDSAvailable)
		return true;		// Already initialised successfully

	HRESULT hr = DirectSoundEnumerate((LPDSENUMCALLBACK)DSEnumProc, NULL);
	if(FAILED(hr))
	{
		if(g_fh) fprintf(g_fh, "DSEnumerate failed (%08X)\n",hr);
		return false;
	}

	if(g_fh)
	{
		fprintf(g_fh, "Number of sound devices = %d\n",num_sound_devices);
	}

	bool bCreatedOK = false;
	for(int x=0; x<num_sound_devices; x++)
	{
		hr = DirectSoundCreate(&sound_device_guid[x], &g_lpDS, NULL);
		if(SUCCEEDED(hr))
		{
			if(g_fh) fprintf(g_fh, "DSCreate succeeded for sound device #%d\n",x);
			bCreatedOK = true;
			break;
		}

		if(g_fh) fprintf(g_fh, "DSCreate failed for sound device #%d (%08X)\n",x,hr);
	}
	if(!bCreatedOK)
	{
		if(g_fh) fprintf(g_fh, "DSCreate failed for all sound devices\n");
		return false;
	}

	hr = g_lpDS->SetCooperativeLevel(framewindow, DSSCL_NORMAL);
	if(FAILED(hr))
	{
		if(g_fh) fprintf(g_fh, "SetCooperativeLevel failed (%08X)\n",hr);
		return false;
	}

#if 0
	DSCAPS DSCaps;
    ZeroMemory(&DSCaps, sizeof(DSBCAPS));
    DSCaps.dwSize = sizeof(DSBCAPS);
	hr = g_lpDS->GetCaps(&DSCaps);
	if(FAILED(hr))
	{
		DirectSound_ErrorText(hr);
		return false;
	}
#endif

	//

	for(DWORD i=0; i<NUM_VOICES; i++)
	{
		hr = DSGetSoundBuffer(&aVoice[i], DSBCAPS_CTRLFREQUENCY | DSBCAPS_CTRLVOLUME);
		if(FAILED(hr))
		{
			if(g_fh) fprintf(g_fh, "Tone %d: DSGetSoundBuffer failed (%08X)\n",i,hr);
			return false;
		}

		hr = DSGetLock(aVoice[i].lpDSBvoice, 0, 0, &pDSLockedBuffer, &dwDSLockedBufferSize, NULL, 0);
		if(FAILED(hr))
		{
			if(g_fh) fprintf(g_fh, "Tone %d: DSGetLock failed (%08X)\n",i,hr);
			return false;
		}

		DSInitToneBuffer(pDSLockedBuffer, i);

		// Commit sound buffer
		hr = aVoice[i].lpDSBvoice->Unlock((void*)pDSLockedBuffer, g_dwDSBufferSize, NULL, 0);
		if(FAILED(hr))
		{
			if(g_fh) fprintf(g_fh, "Tone %d: DSUnlock failed (%08X)\n",i,hr);
			return false;
		}
	}

	//

	for(i=NUM_VOICES; i<NUM_VOICES+NUM_NOISES; i++)
	{
		int nNoiseVoiceNum = i-NUM_VOICES;	// This is also the AY8910 number

		hr = DSGetSoundBuffer(&aVoice[i], DSBCAPS_CTRLPOSITIONNOTIFY | DSBCAPS_CTRLVOLUME);
		if(FAILED(hr))
		{
			if(g_fh) fprintf(g_fh, "Noise %d: DSGetSoundBuffer failed (%08X)\n",i,hr);
			return false;
		}

		hr = DSGetLock(aVoice[i].lpDSBvoice, 0, 0, &pDSLockedBuffer, &dwDSLockedBufferSize, NULL, 0);
		if(FAILED(hr))
		{
			if(g_fh) fprintf(g_fh, "Noise %d: DSGetLock failed (%08X)\n",i,hr);
			return false;
		}

		memset(pDSLockedBuffer, 0x00, g_dwDSBufferSize);

 		hr = aVoice[i].lpDSBvoice->QueryInterface(IID_IDirectSoundNotify, (LPVOID *)&aVoice[i].lpDSNotify);
		if(FAILED(hr))
		{
			if(g_fh) fprintf(g_fh, "Noise %d: QueryInterface failed (%08X)\n",i,hr);
			return false;
		}

		g_hEvents[nNoiseVoiceNum*2+0] = CreateEvent(NULL,	// lpEventAttributes
												FALSE,	// bManualReset (FALSE = auto-reset)
												FALSE,	// bInitialState (FALSE = nonsignaled)
												NULL);	// lpName

		if(g_hEvents[nNoiseVoiceNum*2+0] == NULL)
		{
			if(g_fh) fprintf(g_fh, "Noise %d: CreateEvent0 failed\n",i);
			return false;
		}

		g_hEvents[nNoiseVoiceNum*2+1] = CreateEvent(NULL,	// lpEventAttributes
												FALSE,	// bManualReset (FALSE = auto-reset)
												FALSE,	// bInitialState (FALSE = nonsignaled)
												NULL);	// lpName

		if(g_hEvents[nNoiseVoiceNum*2+1] == NULL)
		{
			if(g_fh) fprintf(g_fh, "Noise %d: CreateEvent1 failed\n",i);
			return false;
		}

		DSBPOSITIONNOTIFY PositionNotify[2];

		PositionNotify[0].dwOffset = g_dwDSBufferSize/2 - 1;	// Half way through buffer
		PositionNotify[0].hEventNotify = g_hEvents[nNoiseVoiceNum+0];

		PositionNotify[1].dwOffset = g_dwDSBufferSize - 1;		// End of buffer
		PositionNotify[1].hEventNotify = g_hEvents[nNoiseVoiceNum+1];

		hr = aVoice[i].lpDSNotify->SetNotificationPositions(2, (DSBPOSITIONNOTIFY*)&PositionNotify);
		if(FAILED(hr))
		{
			if(g_fh) fprintf(g_fh, "Noise %d: SetNotifyPos failed (%08X)\n",i,hr);
			return false;
		}

		// Commit sound buffer
		hr = aVoice[i].lpDSBvoice->Unlock((void*)pDSLockedBuffer, g_dwDSBufferSize, NULL, 0);
		if(FAILED(hr))
		{
			if(g_fh) fprintf(g_fh, "Noise %d: DSUnlock failed (%08X)\n",i,hr);
			return false;
		}
	}

	//

	// Event for terminating thread
	g_hEvents[g_nNumEvents-1] = CreateEvent(NULL,	// lpEventAttributes
											FALSE,	// bManualReset (FALSE = auto-reset)
											FALSE,	// bInitialState (FALSE = nonsignaled)
											NULL);	// lpName

	if(g_hEvents[g_nNumEvents-1] == NULL)
	{
		if(g_fh) fprintf(g_fh, "Thread: CreateEvent failed\n");
		return false;
	}

	//

	DWORD IDThread;

	g_hThread = CreateThread(NULL,
								0, 
								(LPTHREAD_START_ROUTINE) ThreadFunction, 
								&g_hEvents,  // pass event handle
								0,
								&IDThread); 
	if (g_hThread == NULL)
	{
		if(g_fh) fprintf(g_fh, "Thread: CreateThread failed\n");
		return false;
	}

	return true;
}


static void DSUninit()
{
	HRESULT hr;

	if(g_hThread)
	{
		DWORD dwExitCode;
		SetEvent(g_hEvents[g_nNumEvents-1]);	// Signal to thread that it should exit

		do
		{
			if(GetExitCodeThread(g_hThread, &dwExitCode))
			{
				if(dwExitCode == STILL_ACTIVE)
					Sleep(10);
				else
					break;
			}
		}
		while(1);

		CloseHandle(g_hThread);
		g_hThread = NULL;
	}

	for(int i=0; i<NUM_TOTAL_VOICES; i++)
	{
		if(aVoice[i].lpDSBvoice)
		{
			if(aVoice[i].bActive)
			{
				hr = aVoice[i].lpDSBvoice->Stop();
				aVoice[i].bActive = false;
			}
		}

		SAFE_RELEASE(aVoice[i].lpDSNotify);
		SAFE_RELEASE(aVoice[i].lpDSBvoice);
	}

	SAFE_RELEASE(g_lpDS);

	for(i=0; i<g_nNumEvents; i++)
	{
		if(g_hEvents[i])
		{
			CloseHandle(g_hEvents[i]);
			g_hEvents[i] = NULL;
		}
	}
}


static bool PlayVoice(int n, DWORD nFreq, LONG nVolume)
{
	if(!g_bDSAvailable)
		return false;

	HRESULT hr = DS_OK;

	if(n >= NUM_VOICES)
	{
		// Handle noise voices
		if(!aVoice[n].bActive)
		{
			SHORT* pDSLockedBuffer;
			DWORD dwDSLockedBufferSize;

			bool bRes = DSGetLock(aVoice[n].lpDSBvoice, 0, 0, &pDSLockedBuffer, &dwDSLockedBufferSize, NULL, 0);
//			_ASSERT(bRes);
			if(!bRes)
				return false;

			DSInitNoiseBuffer(pDSLockedBuffer, g_dwDSBufferSize, n-NUM_VOICES);

			hr = aVoice[n].lpDSBvoice->Play(0,0,DSBPLAY_LOOPING);
			aVoice[n].bActive = true;

			hr = aVoice[n].lpDSBvoice->Unlock((void*)pDSLockedBuffer, g_dwDSBufferSize, NULL, 0);
			if(FAILED(hr))
				return false;
		}

		return hr == DS_OK ? true : false;
	}

	//

	if(aVoice[n].nFreq != nFreq)
	{
#if NEW_TONE_WAVEFORM
		ULONG nNewFreq = (ULONG) ((double)nFreq * (double)nHalfPeriodSize);
		hr = aVoice[n].lpDSBvoice->SetFrequency(nNewFreq);
		_ASSERT(SUCCEEDED(hr));
#else
		hr = aVoice[n].lpDSBvoice->SetFrequency(nFreq);
#endif
		aVoice[n].nFreq = nFreq;
	}

	if(aVoice[n].nVolume != nVolume)
	{
		hr = aVoice[n].lpDSBvoice->SetVolume(nVolume);
		aVoice[n].nVolume = nVolume;
	}

	if(!aVoice[n].bActive)
	{
		hr = aVoice[n].lpDSBvoice->Play(0,0,DSBPLAY_LOOPING);
		aVoice[n].bActive = true;
	}

#if DBG_MB_STATS
	if(((nFreq < MbStats.nMinFreq) || (MbStats.nMinFreq == 0)) && (nVolume > DSBVOLUME_MIN))
		MbStats.nMinFreq = nFreq;
	if (nFreq > MbStats.nMaxFreq)
		MbStats.nMaxFreq = nFreq;
	for(DWORD i=0,nCnt=0;i<NUM_VOICES;i++)
		if((aVoice[i].bActive) && (aVoice[i].nVolume > DSBVOLUME_MIN))
			nCnt++;
	MbStats.nCurrActiveVoices = nCnt;
	if (nCnt > MbStats.nMaxActiveVoices)
		MbStats.nMaxActiveVoices = nCnt;
#endif

	return hr == DS_OK ? true : false;
}


static bool StopVoice(int n)
{
	if(!g_bDSAvailable)
		return false;

	HRESULT hr = DS_OK;

	if(aVoice[n].bActive)
	{
		hr = aVoice[n].lpDSBvoice->Stop();
		aVoice[n].bActive = false;
	}

	return hr == DS_OK ? true : false;
}

//
// ----- ALL GLOBALLY ACCESSIBLE FUNCTIONS ARE BELOW THIS LINE -----
//

//===========================================================================

void MB_Initialize()
{
	memset(&g_MB,0,sizeof(g_MB));
	
	for(int i=0; i<NUM_MB*NUM_DEVS_PER_MB; i++)
		g_MB[i].nAY8910Number = i;

	//

	for(i=0; i<=15; i++)
	{
		aVolTbl[i] = (LONG) (-(exp((15-i)*log((-DSBVOLUME_MIN)+1)/15) + 0.5 - 1.0));
	}

	//

	g_bDSAvailable = DSInit();

	QueryPerformanceFrequency((LARGE_INTEGER*)&g_nPerfFreq);
	if(g_fh) fprintf(g_fh, "Performance frequency = %d\n",g_nPerfFreq);
}

void MB_Destroy()
{
	DSUninit();
	g_bDSAvailable = false;
}

void MB_Reset()
{
	if(!g_bDSAvailable)
		return;

	for(int i=0; i<NUM_MB*NUM_DEVS_PER_MB; i++)
	{
		ResetSY6522(&g_MB[i]);
		ResetAY8910(&g_MB[i]);
	}
}

BYTE MB_Read(WORD nAddr)
{
	if(!g_bMBEnable)
		return 0;

	BYTE nMB = (nAddr>>8)&0xf - SLOT4;
	BYTE nOffset = nAddr&0xff;

	if(nOffset <= (SY6522A_Offset+0x0F))
		return SY6522_Read(nMB*NUM_DEVS_PER_MB + SY6522_DEVICE_A, nAddr&0xf);
	else if((nOffset >= SY6522B_Offset) && (nOffset <= (SY6522B_Offset+0x0F)))
		return SY6522_Read(nMB*NUM_DEVS_PER_MB + SY6522_DEVICE_B, nAddr&0xf);
	else
		return 0;
}

void MB_Write(WORD nAddr, BYTE nValue)
{
	if(!g_bMBEnable)
		return;

	BYTE nMB = (nAddr>>8)&0xf - SLOT4;
	BYTE nOffset = nAddr&0xff;

	if(nOffset <= (SY6522A_Offset+0x0F))
		SY6522_Write(nMB*NUM_DEVS_PER_MB + SY6522_DEVICE_A, nAddr&0xf, nValue);
	else if((nOffset >= SY6522B_Offset) && (nOffset <= (SY6522B_Offset+0x0F)))
		SY6522_Write(nMB*NUM_DEVS_PER_MB + SY6522_DEVICE_B, nAddr&0xf, nValue);
}

void MB_Mute()
{
	if(!g_bMBEnable)
		return;

	for(int i=0; i<NUM_TOTAL_VOICES; i++)
		if(aVoice[i].bActive)
			aVoice[i].lpDSBvoice->SetVolume(DSBVOLUME_MIN);
}

void MB_Demute()
{
	if(!g_bMBEnable)
		return;

	for(int i=0; i<NUM_TOTAL_VOICES; i++)
		if(aVoice[i].bActive)
			aVoice[i].lpDSBvoice->SetVolume(aVoice[i].nVolume);
}

void MB_CheckIRQ()
{
	if(!g_bMBEnable)
		return;

	static ULONGLONG nLastPerfCnt = 0;

	if(!g_nPerfFreq)
		return;

	if(g_bMBTimerIrqActive && g_n6522TimerPeriod)
	{
		DWORD nIrqCnt = (DWORD) ((double)g_n6522TimerPeriod / CLK_6502 * (double)(LONGLONG)g_nPerfFreq);

		bool bDoCpuIrq = false;
		ULONGLONG nCurrPerfCnt;
		QueryPerformanceCounter((LARGE_INTEGER*)&nCurrPerfCnt);
		if((nCurrPerfCnt - nLastPerfCnt) > nIrqCnt)	// PerfCnt won't wrap in our life time!
			bDoCpuIrq = true;

		if(bDoCpuIrq)
		{
			g_MB[g_nMBTimerDevice].sy6522.IFR = TIMER1;
			CpuIRQ();
			nLastPerfCnt = nCurrPerfCnt;

			if((g_MB[g_nMBTimerDevice].sy6522.ACR & RUNMODE) == RM_ONESHOT)
			{
				// One-shot mode
				StopTimer(&g_MB[g_nMBTimerDevice]);
			}
			else
			{
				// Free-running mode
				// - Ultima4/5 change ACCESS_TIMER1 after a couple of IRQs into tune
				g_MB[g_nMBTimerDevice].sy6522.TIMER1L = g_MB[g_nMBTimerDevice].sy6522.ACCESS_TIMER1L;
				g_MB[g_nMBTimerDevice].sy6522.TIMER1H = g_MB[g_nMBTimerDevice].sy6522.ACCESS_TIMER1H;
				StartTimer(&g_MB[g_nMBTimerDevice]);
			}
		}
	}
}

void MB_ToggleEnaDis()
{
	if(g_bMBEnable)
	{
		MB_Mute();
		g_bMBEnable = false;
	}
	else
	{
		g_bMBEnable = true;
	}
}
